#version 450 core

in vec3 worldPosFrag;
in vec2 texCoords;
in mat3 TBN;

uniform float far_plane_camera;
uniform float near_plane_camera;

uniform sampler2D texture_diffuse1;
uniform sampler3D volFogVoxelGrid;

uniform mat4 viewProjection;

layout(std140, binding = 0) uniform VP
{
    mat4 view;
    mat4 projection;
    vec4 waterClipPlane;
};


layout(std140, binding = 2) uniform UserInput
{
    ivec4 volFogAct_blueNoise_filterAct;
};

out vec4 FragColor;


float exponential2LinearDepth(float z)
{
    float ratioFarNear = far_plane_camera / near_plane_camera;
    return 1.0f / ((1.0f - ratioFarNear) * z + ratioFarNear);
}


// Converts world space coordinates to ndc coordinates
vec3 ws2ndc(vec3 worldPos, mat4 vp)
{
    vec4 p = vp * vec4(worldPos, 1.0f);
    return p.xyz / p.w;
}


// Converts ndc coordinates to uv coordinates
vec3 ndc2uvw(vec3 ndc)
{
    vec3 uvw = vec3(
		ndc.x * 0.5f + 0.5f,
		ndc.y * 0.5f + 0.5f,
		exponential2LinearDepth(ndc.z * 0.5f + 0.5f)
	);

    // Exponential z
    vec2 params = vec2(float(128) / log2(far_plane_camera / near_plane_camera), -(float(128) * log2(near_plane_camera) / log2(far_plane_camera / near_plane_camera)));

    float view_z = uvw.z * far_plane_camera;
    uvw.z = (max(log2(view_z) * params.x + params.y, 0.0f)) / 128;
    return uvw;
}

// Converts world space coordinates to ndc coordinates
vec3 ws2uvw(vec3 worldPos, mat4 vp)
{
    vec3 ndc = ws2ndc(worldPos, vp);
    return ndc2uvw(ndc);
}

#define VOXEL_GRID_SIZE_X 160
#define VOXEL_GRID_SIZE_Y 90
#define VOXEL_GRID_SIZE_Z 128

float gauss(int offset, float sigma) {
    return exp(-(offset * offset) / (2.0 * sigma * sigma)) / (sqrt(2.0 * 3.14159265) * sigma);
}


vec4 gauss_1d(float sigma, vec3 uvw, int axis, int axisGridSize) {
    float weightSum = 0.0f;
    vec4 result = vec4(0.0f, 0.0f, 0.0f, 0.0f);
    for (int i = -2; i < 3; i++) {
        float newAxisValue = (uvw[axis] * axisGridSize + i) / axisGridSize;
		
		// Ensure we don't go out of bounds
		if (newAxisValue < 0.0f)
            newAxisValue = 0.0f;
        else if (newAxisValue > 1.0f)
            newAxisValue = 1.0f;

        uvw[axis] = newAxisValue;

        float w = gauss(i, sigma);
        weightSum += w;
        result += w * textureLod(volFogVoxelGrid, uvw, 0.0f);
    }
    return result / weightSum;
}


vec4 blur(vec3 uvw) {
    float sigma = 1;
    
	// Blur in each direction seperately -> way faster than applying a 3D kernel.
    vec4 xResult = gauss_1d(sigma, uvw, 0, VOXEL_GRID_SIZE_X);
    vec4 yResult = gauss_1d(sigma, uvw, 1, VOXEL_GRID_SIZE_Y);
    vec4 zResult = gauss_1d(sigma, uvw, 2, VOXEL_GRID_SIZE_Z);

    return (xResult + yResult + zResult) / 3.0f;
}


vec3 applyVolumetricFog(vec3 color, vec3 worldPos)
{
    vec3 uvw = ws2uvw(worldPos, viewProjection);
    vec4  volFogLight = blur(uvw);

	float transmittance   = volFogLight.a;
    //return color * transmittance + volFogLight.rgb;
	return color * max(volFogLight.a, 0.6f) + volFogLight.rgb * 0.3f;
}


void main() {
	vec3 color = texture(texture_diffuse1, texCoords).xyz * 1.2;
	if (volFogAct_blueNoise_filterAct.x == 1)
		color = applyVolumetricFog(color, worldPosFrag);
	FragColor = vec4(color, 1.0f);
}