#version 450 core

in vec3 worldPosFrag;
in vec2 texCoords;
in mat3 TBN;

uniform mat4 model;
uniform vec3 cameraPos;
uniform float far_plane_light;
uniform float far_plane_camera;
uniform float near_plane_camera;

uniform sampler2D texture_diffuse1;
uniform sampler2D texture_specular1;
uniform sampler2D texture_normal1;
uniform sampler2D texture_alpha1;
uniform sampler2D texture_glow1;
uniform samplerCubeArray depthMaps;

uniform bool hasAlphaTexture;
uniform bool hasGlowTexture;
uniform bool hasNormalTexture;
uniform bool tesselationEnabled;
uniform bool volFogEnable;
uniform bool moonDisabled;

uniform sampler3D volFogVoxelGrid;

layout(std140, binding=1) uniform Lights{
	int numLights;
	vec4 position[11];
	vec3 ambient[11];
	vec3 diffuse[11];
	vec3 specular[11];
	vec3 attenuation[11];
	float lightFar[11];
};


layout(std140, binding = 0) uniform VP
{
    mat4 view;
    mat4 projection;
    vec4 waterClipPlane;
};


layout(std140, binding = 2) uniform UserInput
{
    ivec4 volFogAct_blueNoise_filterAct;
};



out vec4 FragColor;


float exponential2LinearDepth(float z)
{
    float ratioFarNear = far_plane_camera / near_plane_camera;
    return 1.0f / ((1.0f - ratioFarNear) * z + ratioFarNear);
}


// Converts world space coordinates to ndc coordinates
vec3 ws2ndc(vec3 worldPos, mat4 viewProjection)
{
    vec4 p = viewProjection * vec4(worldPos, 1.0f);
    return p.xyz / p.w;
}


// Converts ndc coordinates to uv coordinates
vec3 ndc2uvw(vec3 ndc)
{
    vec3 uvw = vec3(
		ndc.x * 0.5f + 0.5f,
		ndc.y * 0.5f + 0.5f,
		exponential2LinearDepth(ndc.z * 0.5f + 0.5f)
	);

    // Exponential z
    vec2 params = vec2(float(128) / log2(far_plane_camera / near_plane_camera), -(float(128) * log2(near_plane_camera) / log2(far_plane_camera / near_plane_camera)));

    float view_z = uvw.z * far_plane_camera;
    uvw.z = (max(log2(view_z) * params.x + params.y, 0.0f)) / 128;
    return uvw;
}

// Converts world space coordinates to ndc coordinates
vec3 ws2uvw(vec3 worldPos, mat4 vp)
{
    vec3 ndc = ws2ndc(worldPos, vp);
    return ndc2uvw(ndc);
}

#define VOXEL_GRID_SIZE_X 160
#define VOXEL_GRID_SIZE_Y 90
#define VOXEL_GRID_SIZE_Z 128

float gauss(int offset, float sigma) {
    return exp(-(offset * offset) / (2.0 * sigma * sigma)) / (sqrt(2.0 * 3.14159265) * sigma);
}


vec4 gauss_1d(float sigma, vec3 uvw, int axis, int axisGridSize) {
    float weightSum = 0.0f;
    vec4 result = vec4(0.0f, 0.0f, 0.0f, 0.0f);
    for (int i = -2; i < 3; i++) {
        float newAxisValue = (uvw[axis] * axisGridSize + i) / axisGridSize;
		
		// Ensure we don't go out of bounds
		if (newAxisValue < 0.0f)
            newAxisValue = 0.0f;
        else if (newAxisValue > 1.0f)
            newAxisValue = 1.0f;

        uvw[axis] = newAxisValue;

        float w = gauss(i, sigma);
        weightSum += w;
        result += w * textureLod(volFogVoxelGrid, uvw, 0.0f);
    }
    return result / weightSum;
}


vec4 blur(vec3 uvw) {
    float sigma = 1;
    
	// Blur in each direction seperately -> way faster than applying a 3D kernel.
    vec4 xResult = gauss_1d(sigma, uvw, 0, VOXEL_GRID_SIZE_X);
    vec4 yResult = gauss_1d(sigma, uvw, 1, VOXEL_GRID_SIZE_Y);
    vec4 zResult = gauss_1d(sigma, uvw, 2, VOXEL_GRID_SIZE_Z);

    return (xResult + yResult + zResult) / 3.0f;
}


vec3 applyVolumetricFog(vec3 color, vec3 worldPos)
{
    vec3 uvw = ws2uvw(worldPos, projection * view);

    vec4  volFogLight = volFogAct_blueNoise_filterAct.z == 1 ? blur(uvw) : textureLod(volFogVoxelGrid, uvw, 0.0f);
    
	float transmittance   = volFogLight.a;

    return color * transmittance + volFogLight.rgb;
	//return color * max(volFogLight.a, 0.7f) + volFogLight.rgb * 0.2f;
}


void main() {
	float alphaValue = 1.0f;
	if (hasAlphaTexture) {
		// As our models don't contain semitransparent elements, it is okay to just discard the whole fragment.
		alphaValue = texture(texture_alpha1, texCoords).r;
		if (alphaValue < 0.01)
			discard;
	}

	vec3 glow = vec3(0.0f);
	if (hasGlowTexture)
		glow = texture(texture_glow1, texCoords).rgb;

	//Lighting calculations in world space

	mat3 newTBN = TBN;
	//Tangent Space Calculation
	vec3 normal;
	if (hasNormalTexture) {
		normal = texture(texture_normal1, texCoords).rgb;
		normal = normalize(normal * 2.0f - 1.0f);
		normal = normalize(TBN * normal);
		newTBN = TBN;
	}
	else {
		normal = TBN[2];
		// TODO: check why tesselation inverts triangle facing
		if (tesselationEnabled) {
			if (gl_FrontFacing)
				normal = -normal;
			}
		else {
			if (!gl_FrontFacing)
				normal = -normal;
		}
		newTBN[2] = normal;
	}

	vec3 moonAmbient = vec3(0);
	if (moonDisabled)
		moonAmbient = vec3(0.05f) * texture(texture_diffuse1, texCoords).xyz;
	vec3 result = moonAmbient + glow;
	for(int i = 0;i < numLights;i++){
		
		//point light
		if(position[i].w == 1){
			
			//AMBIENT
			vec3 amb = ambient[i] * texture(texture_diffuse1, texCoords).xyz;

			//DIFFUSE
			vec3 fragToLight =  normalize(position[i].xyz - worldPosFrag);
			float cosine = max(dot(normal, fragToLight), 0.0f);
			vec3 diff = (diffuse[i] * cosine) * texture(texture_diffuse1, texCoords).xyz;

			//SPECULAR
			vec3 fragToEye =normalize(cameraPos - worldPosFrag);
			vec3 halfway = normalize(fragToLight + fragToEye);

			cosine = max(dot(normal, halfway), 0.0f);
			vec3 spec = (specular[i] * pow(cosine, 20.0f)) * texture(texture_specular1, texCoords).xyz;

			//ATTENUATION
			// saftey measure for avoiding division by zero
			float dist = max(abs(length(position[i].xyz - worldPosFrag)), 0.000001);
			float att = 1.0 / (attenuation[i][0] + attenuation[i][1]*dist + attenuation[i][2]* (dist*dist));

			//SHADOW CALCULATIONS
			// use simple Percentage Closer Filtering that samples multiple values from the depth map and averages them, this leads to softer shadows
			// the sampling is done by applying small perturbations to the texture sampling vector
			// the perturbation factor describes the "softness" of the lightsource
			// to hit the right factor heavily depends on the distances in the scene, the one that we choose here might look worse in another scene
			// theoretically, the calculation of currentDepth using the perturbed vector is slightly wrong and makes the assumption that nearby fragments
			// have a similar depth, but it works fine in our case

			float shadow = 0.0;

			for(int x = -1; x <= 1; ++x)
			{
				for(int y = -1; y <= 1; ++y)
				{
					for(int z = -1; z <=1; ++z){
						vec3 lightToFrag = worldPosFrag - position[i].xyz;

						float perturbationFactor = 0.02;
						vec3 perturbationVector = vec3(x,y,z)*perturbationFactor;
						vec3 lightToFragPerturbed = lightToFrag + perturbationVector;

						float closestDepth = texture(depthMaps, vec4(lightToFragPerturbed, i)).r;
						closestDepth *= lightFar[i];
						float currentDepth = length(lightToFragPerturbed);
						float bias = 0.05;
						shadow += (currentDepth -  bias) > closestDepth ? 1.0f : 0.0f;
						}
					}
			}
			shadow /= 27.0;
			//att = 1.0f;
			result += (amb + vec3((1.0-shadow)*(diff+spec).x,(1.0-shadow)*(diff+spec).y,(1.0-shadow)*(diff+spec).z))*att;
			}
		}
	// Volumetric Light
	if (volFogAct_blueNoise_filterAct.x == 1 && volFogEnable)
		result = applyVolumetricFog(result, worldPosFrag);

	FragColor = vec4(result, alphaValue);
}