#version 450 core

struct PatchData
{
	vec3 b300;
	vec3 b030;
	vec3 b003;
	vec3 b210;
	vec3 b120;
	vec3 b021;
	vec3 b012;
	vec3 b102;
	vec3 b201;
	vec3 b111;

	vec3 n200;
	vec3 n020;
	vec3 n002;
	vec3 n110;
	vec3 n011;
	vec3 n101;
};

layout(vertices = 3) out;

uniform mat4 modelMatrix;
uniform mat4 viewMatrix;

in VertexData {
	vec3 positionOS;    // vertex position in object space
	vec2 texCoords;     // texture coordinates
	vec3 normalOS;      // normal in object space
	vec3 tangentOS;     // tangent in object space
	vec3 bitangentOS;   // bitangent in object space
} tc_in[];

out TescTeseData
{
	vec2 texCoords;     // texture coordinates
	vec3 tangentOS;     // tangent in object space
	vec3 bitangentOS;   // bitangent in object space
} tc_out[];

layout (location = 4) patch out PatchData pd;

vec3 projectPointToPlane(vec3 point, vec3 planePoint, vec3 planeNormal)
{
	vec3 v = point - planePoint;
	float len = dot(v, planeNormal);
	vec3 d = len * planeNormal;
	return point - d;
}

vec3 mirrorVecAtPlane(vec3 vector, vec3 planeNormal)
{
	planeNormal = normalize(planeNormal);
	float len = dot(vector, planeNormal);
	vec3 d = (2 * len) * planeNormal;
	return vector - d;
}

vec3 interpolate2(float f1, float f2, float f3, vec3 P1, vec3 P2, vec3 P3)
{
	return (f1 * P1 + f2 * P2 + f3 * P3) / 2;
}

vec3 interpolate3(float f1, float f2, float f3, vec3 P1, vec3 P2, vec3 P3)
{
	return (f1 * P1 + f2 * P2 + f3 * P3) / 3;
}

void main()
{
	if (gl_InvocationID < 3)
	{
		tc_out[gl_InvocationID].texCoords = tc_in[gl_InvocationID].texCoords;
		tc_out[gl_InvocationID].tangentOS = tc_in[gl_InvocationID].tangentOS;
		tc_out[gl_InvocationID].bitangentOS = tc_in[gl_InvocationID].bitangentOS;
	}
	
	if (gl_InvocationID == 0)
	{
		mat4 itMV = inverse(transpose(viewMatrix * modelMatrix));
		vec4 normalVS1 = itMV * vec4(tc_in[0].normalOS, 1.0);
		vec4 normalVS2 = itMV * vec4(tc_in[1].normalOS, 1.0);
		vec4 normalVS3 = itMV * vec4(tc_in[2].normalOS, 1.0);

		const float userDefinedTessLevel = 5;
		const float tessLevel =
			1.0 + // A tess level of 0.0 would mean that the patch is culled. Generally, we want to render patches => therefore, 1.0.
			(userDefinedTessLevel - 1.0); // The patches to be tessellated might get higher tess levels applied.

		gl_TessLevelOuter[0] = tessLevel;
		gl_TessLevelOuter[1] = tessLevel;
		gl_TessLevelOuter[2] = tessLevel;
		gl_TessLevelInner[0] = tessLevel;

		vec3 P1 = tc_in[0].positionOS;
		vec3 P2 = tc_in[1].positionOS;
		vec3 P3 = tc_in[2].positionOS;
		vec3 N1 = tc_in[0].normalOS;
		vec3 N2 = tc_in[1].normalOS;
		vec3 N3 = tc_in[2].normalOS;

		vec3 m300 = interpolate3(3, 0, 0,   P1, P2, P3);
		vec3 m030 = interpolate3(0, 3, 0,   P1, P2, P3);
		vec3 m003 = interpolate3(0, 0, 3,   P1, P2, P3);
		vec3 m021 = interpolate3(0, 2, 1,   P1, P2, P3);
		vec3 m012 = interpolate3(0, 1, 2,   P1, P2, P3);
		vec3 m201 = interpolate3(2, 0, 1,   P1, P2, P3);
		vec3 m102 = interpolate3(1, 0, 2,   P1, P2, P3);
		vec3 m210 = interpolate3(2, 1, 0,   P1, P2, P3);
		vec3 m120 = interpolate3(1, 2, 0,   P1, P2, P3);
		vec3 m111 = interpolate3(1, 1, 1,   P1, P2, P3);

		vec3 b300 = m300;
		vec3 b030 = m030;
		vec3 b003 = m003;
		vec3 b021 = projectPointToPlane(m021, P2, N2);
		vec3 b012 = projectPointToPlane(m012, P3, N3);
		vec3 b102 = projectPointToPlane(m102, P3, N3);
		vec3 b201 = projectPointToPlane(m201, P1, N1);
		vec3 b210 = projectPointToPlane(m210, P1, N1);
		vec3 b120 = projectPointToPlane(m120, P2, N2);
		vec3 E = (b021 + b012 + b102 + b201 + b210 + b120) / 6;
		vec3 V = (P1 + P2 + P3) / 3;
		vec3 b111 = E + (E - V) / 2;

		pd.b030 = b030;
		pd.b003 = b003;
		pd.b300 = b300;
		pd.b021 = b021;
		pd.b012 = b012;
		pd.b102 = b102;
		pd.b201 = b201;
		pd.b210 = b210;
		pd.b120 = b120;
		pd.b111 = b111;

		vec3 m200 = interpolate2(2, 0, 0,   N1, N2, N3);
		vec3 m020 = interpolate2(0, 2, 0,   N1, N2, N3);
		vec3 m002 = interpolate2(0, 0, 2,   N1, N2, N3);
		vec3 m011 = interpolate2(0, 1, 1,   N1, N2, N3);
		vec3 m101 = interpolate2(1, 0, 1,   N1, N2, N3);
		vec3 m110 = interpolate2(1, 1, 0,   N1, N2, N3);

		vec3 n200 = m200;
		vec3 n020 = m020;
		vec3 n002 = m002;
		vec3 n011 = mirrorVecAtPlane(m011, P3 - P2); 
		vec3 n101 = mirrorVecAtPlane(m101, P3 - P1); 
		vec3 n110 = mirrorVecAtPlane(m110, P2 - P1); 

		pd.n200 = normalize(n200);
		pd.n020 = normalize(n020);
		pd.n002 = normalize(n002);
		pd.n011 = normalize(n011);
		pd.n101 = normalize(n101);
		pd.n110 = normalize(n110);

	}
}
