#version 450 core 

out vec4 FragColor;
  
in vec4 clipSpace;
in vec2 texCoords;
in vec3 worldPosFrag;

uniform sampler2D reflectionTexture;
uniform sampler2D refractionTexture;
uniform sampler2D dudv;
uniform sampler2D normal;
uniform float waveFactor;
uniform vec3 cameraPos;

uniform sampler3D volFogVoxelGrid;
uniform float far_plane_camera;
uniform float near_plane_camera;


layout(std140, binding = 0) uniform VP
{
    mat4 view;
    mat4 projection;
    vec4 waterClipPlane;
};

layout(std140, binding=1) uniform Lights{
	int numLights;
	vec4 position[11];
	vec3 ambient[11];
	vec3 diffuse[11];
	vec3 specular[11];
	vec3 attenuation[11];
	float lightFar[11];
};


layout(std140, binding = 2) uniform UserInput
{
    ivec4 volFogAct_blueNoise_filterAct;
};



float exponential2LinearDepth(float z)
{
    float ratioFarNear = far_plane_camera / near_plane_camera;
    return 1.0f / ((1.0f - ratioFarNear) * z + ratioFarNear);
}


// Converts world space coordinates to ndc coordinates
vec3 ws2ndc(vec3 worldPos, mat4 viewProjection)
{
    vec4 p = viewProjection * vec4(worldPos, 1.0f);
    return p.xyz / p.w;
}


// Converts ndc coordinates to uv coordinates
vec3 ndc2uvw(vec3 ndc)
{
    vec3 uvw = vec3(
		ndc.x * 0.5f + 0.5f,
		ndc.y * 0.5f + 0.5f,
		exponential2LinearDepth(ndc.z * 0.5f + 0.5f)
	);

    // Exponential z
    vec2 params = vec2(float(128) / log2(far_plane_camera / near_plane_camera), -(float(128) * log2(near_plane_camera) / log2(far_plane_camera / near_plane_camera)));

    float view_z = uvw.z * far_plane_camera;
    uvw.z = (max(log2(view_z) * params.x + params.y, 0.0f)) / 128;
    return uvw;
}

// Converts world space coordinates to ndc coordinates
vec3 ws2uvw(vec3 worldPos, mat4 vp)
{
    vec3 ndc = ws2ndc(worldPos, vp);
    return ndc2uvw(ndc);
}

float gauss(int offset, float sigma) {
    return exp(-(offset * offset) / (2.0 * sigma * sigma)) / (sqrt(2.0 * 3.14159265) * sigma);
}


vec4 gauss_1d(float sigma, vec3 uvw, int axis, int axisGridSize) {
    float weightSum = 0.0f;
    vec4 result = vec4(0.0f, 0.0f, 0.0f, 0.0f);
    for (int i = -2; i < 3; i++) {
        float newAxisValue = (uvw[axis] * axisGridSize + i) / axisGridSize;
		
		// Ensure we don't go out of bounds
		if (newAxisValue < 0.0f)
            newAxisValue = 0.0f;
        else if (newAxisValue > 1.0f)
            newAxisValue = 1.0f;

        uvw[axis] = newAxisValue;

        float w = gauss(i, sigma);
        weightSum += w;
        result += w * textureLod(volFogVoxelGrid, uvw, 0.0f);
    }
    return result / weightSum;
}


#define VOXEL_GRID_SIZE_X 160
#define VOXEL_GRID_SIZE_Y 90
#define VOXEL_GRID_SIZE_Z 128
vec4 blur(vec3 uvw) {
    float sigma = 1;
    
	// Blur in each direction seperately -> way faster than applying a 3D kernel.
    vec4 xResult = gauss_1d(sigma, uvw, 0, VOXEL_GRID_SIZE_X);
    vec4 yResult = gauss_1d(sigma, uvw, 1, VOXEL_GRID_SIZE_Y);
    vec4 zResult = gauss_1d(sigma, uvw, 2, VOXEL_GRID_SIZE_Z);

    return (xResult + yResult + zResult) / 3.0f;
}

vec3 applyVolumetricFog(vec3 color, vec3 worldPos)
{
    vec3 uvw = ws2uvw(worldPos, projection * view);

    vec4  volFogLight = volFogAct_blueNoise_filterAct.z == 1 ? blur(uvw) : textureLod(volFogVoxelGrid, uvw, 0.0f);
    
	float transmittance   = volFogLight.a;

    return color * transmittance + volFogLight.rgb;
	//return color * max(volFogLight.a, 0.7f) + volFogLight.rgb * 0.2f;
}

void main()
{
    float tiling = 500.0; //tiling the dudv texture for "bigger" resolution
    float strength = 0.05; //strngth of the wave distortion
    float waveSpeed = 0.0000001; //speed of the waves

    //fresnel effect (the higher the angle from the camera to the surface of the water, the more transpartent the water becomes)
    //we sample the texture with rbg instead of rgb because the normals are given in tangent space
    //and since we are rotating the quad the water gets rendere on to lay flat, we switch the y and z components so the normals are correct again
    vec3 surfaceNormal = vec3(0,1,0);
    vec3 fragToEye = normalize(cameraPos - worldPosFrag);
    float fresnelCosine = max(0.0, dot(surfaceNormal, fragToEye));

    //specular highlights
    vec3 spec = vec3(0,0,0);
    for(int light = 0; light < numLights; light++)
    {
    vec3 specularNormal = texture(normal, (texCoords.xy + vec2(waveFactor * waveSpeed))*tiling).rbg;
	specularNormal = normalize(specularNormal *2.0 -1.0);
    vec3 fragToLight = normalize(position[light].xyz - worldPosFrag);
    vec3 halfway = normalize(fragToLight + fragToEye);
   	float specularCosine = max(dot(specularNormal, halfway), 0.0f);

	float dist = length(position[light].xyz - worldPosFrag);
	float att = 1.0 / (attenuation[light][0] + attenuation[light][1]*dist + attenuation[light][2]* (dist*dist));
		
	spec += att * (specular[light] * pow(specularCosine, 30.0f));
    }


    vec2 ndc = (clipSpace.xy/clipSpace.w)/2.0 + 0.5;



    //the tiling lets the dudv texture be samples like it's tiled across the quad
    //strenght determines the strength of the dudv distorion
    //waveFactor * waveSpeed adds an additional offset that constantly goes from 0 to 1 and lets the dudv texture "slide" across the quad periodically
    vec2 distortion = ((texture(dudv, (texCoords.xy + vec2(waveFactor * waveSpeed))*tiling)).xy * 2.0 - 1.0)*strength;

    //have to clamp the texture coordinates because the distortion would cause them
    //to "overflow" on the edges
    vec2 refractionTexCoords = ndc.xy + distortion;
    refractionTexCoords = clamp(refractionTexCoords, 0.001, 0.999);

    vec2 reflectionTexCoords = vec2(ndc.x, -ndc.y) + distortion;
    reflectionTexCoords.x = clamp(reflectionTexCoords.x, 0.001, 0.999);
    reflectionTexCoords.y = clamp(reflectionTexCoords.y, -0.999, -0.001);

    vec3 reflectionColor = texture(reflectionTexture, reflectionTexCoords).xyz;
    vec3 refractionColor = texture(refractionTexture, refractionTexCoords).xyz;


    vec3 result = fresnelCosine * refractionColor + (1-fresnelCosine) * reflectionColor;
    result = vec3(result.r * 0.85, result.g * 0.9, result.b);
    result += spec;
	
	if (volFogAct_blueNoise_filterAct.x == 1)
		result = applyVolumetricFog(result, worldPosFrag);

    FragColor = vec4(result,1.0f);

} 